package ca.recrown.islandsurvivalcraft.interaction.commands.runnables;

import java.util.HashSet;
import java.util.LinkedList;
import java.util.concurrent.CompletableFuture;

import org.bukkit.Bukkit;
import org.bukkit.Color;
import org.bukkit.Location;
import org.bukkit.Particle;
import org.bukkit.World;
import org.bukkit.command.CommandSender;
import org.bukkit.entity.Player;

import ca.recrown.islandsurvivalcraft.IslandSurvivalCraft;
import ca.recrown.islandsurvivalcraft.utilities.datatypes.Point2;
import ca.recrown.islandsurvivalcraft.utilities.pathfinding.DepthFirstSearch;
import ca.recrown.islandsurvivalcraft.world.IslandWorldMap;
import ca.recrown.islandsurvivalcraft.world.WorldInfo;

public class HighlightIslandCommand implements CommandRunnable {
    private final Particle.DustOptions OPTIONS = new Particle.DustOptions(Color.RED, 1f);
    private final int PARTICLE_AMOUNT = 3;
    private final double OFFSET_X = 0.0d, OFFSET_Y = 6.0d, OFFSET_Z = 0.0d;
    private final double EXTRA = 0d;
    HashSet<Player> playersHighlighting = new HashSet<>();
    LinkedList<PlayerIslandHighlighter> waitingList = new LinkedList<>();

    @Override
    public String getDescription() {
        return "Displays particle effects at edge of islands.";
    }

    @Override
    public CompletableFuture<String> invoke(CommandSender sender, String[] args) {
        CompletableFuture<String> completableFuture = new CompletableFuture<>();
        if (!(sender instanceof Player)) {
            completableFuture.complete("This command can only be run as a player!");
            return completableFuture;
        }
        Player player = (Player) sender;
        if (args[0].isEmpty()) {
            completableFuture.complete("Arguments required. Please refer to \"/IslandSurvivalCraft help\" for more info.");
            return completableFuture;
        }
        if (args[0].toLowerCase().equals("start")) {
            if (playersHighlighting.add(player)) {
                waitingList.add(new PlayerIslandHighlighter(player));
                completableFuture.complete("You are now highlighting islands.");
            } else {
                completableFuture.complete("You are already highlighting islands.");
            }
        } else if (args[0].toLowerCase().equals("stop")) {
            if (playersHighlighting.remove(player)) {
                completableFuture.complete("You are no longer highlighting islands.");
            } else {
                completableFuture.complete("You weren't highlighting islands.");
            }
        } else {
            completableFuture.complete(String.format("Argument \"%s\" was not understood. Please refer to \"/IslandSurvivalCraft help\" for more info.", args[0]));
        }
        return completableFuture;
    }

    @Override
    public void initialize(IslandSurvivalCraft islandsurvivalcraft) {        
        Bukkit.getScheduler().scheduleSyncRepeatingTask(islandsurvivalcraft, new Runnable() {
            @Override
            public void run() {
                PlayerIslandHighlighter highlighter = waitingList.poll();
                if (highlighter == null) return;
                Player player = highlighter.player;
                if (player == null || !player.isOnline() || !playersHighlighting.contains(player)) {
                    playersHighlighting.remove(player);
                    return;
                }
                World world = player.getWorld();
                WorldInfo worldInfo = islandsurvivalcraft.getWorldInfoManager().retrieve(world);
                Location playerLocation = player.getLocation();
                int playerX = playerLocation.getBlockX(), playerY = playerLocation.getBlockY(), playerZ = playerLocation.getBlockZ();
                IslandWorldMap map = worldInfo.getIslandMap();
                
                for (int i = 0; i < 8; i++) {
                    highlighter.update(new Point2(playerX, playerZ), map);
                }

                LinkedList<Point2> highlighted = highlighter.getHighlighted();
                for (Point2 point2 : highlighted) {
                    if (point2 != null) {
                        spawnParticle(world, point2.x, playerY, point2.y);
                    }
                }
                waitingList.add(highlighter);
            }
        }, 0, 8);

    }

    private void spawnParticle(World world, double x, double y, double z) {
        world.spawnParticle(Particle.REDSTONE, x, y, z, PARTICLE_AMOUNT, OFFSET_X, OFFSET_Y, OFFSET_Z, EXTRA, OPTIONS, true);
    }

    @Override
    public String getDetailedDescription() {
        return "A helpful debugging command to analyze what the plugin considers an island by displaying particle effects around the player that is a part of the island. The <start | stop> argument is required to dicate whether or not to start the highlighting or stop it.";
    }
    
    private class PlayerIslandHighlighter {
        public final int TRAIL_LENGTH = 64;
        private final int MAX_SEARCH = 20;
        private final int MAX_SEARCH_DIST = 16;
        private final Player player;
        private final LinkedList<Point2> highlighted = new LinkedList<>();
        private DepthFirstSearch dfs;

        public PlayerIslandHighlighter(Player player) {
            this.player = player;
        }

        private void addHighlight(Point2 point) {
            highlighted.add(point);
            if (highlighted.size() > TRAIL_LENGTH) {
                highlighted.poll();
            }
        }

        public void update(Point2 playerLocation, IslandWorldMap islandWorldMap) {
            if (dfs == null || dfs.step(islandWorldMap.islandValidator, (p) -> p.distance(playerLocation) > MAX_SEARCH_DIST)) {
                dfs = new DepthFirstSearch(playerLocation, null, (p1, p2) -> {
                    double diff = islandWorldMap.getWorldValue(p1.x, p1.y) - islandWorldMap.getWorldValue(p2.x, p2.y);
                    if (diff == 0) {
                        return 0;
                    } else if (diff > 0) {
                        return 1;
                    } else {
                        return -1;
                    }
                }, MAX_SEARCH);
            }
            Point2 currentPoint = dfs.getCurrentNode();
            if (currentPoint != null && islandWorldMap.isEdgeOfIsland(currentPoint.x, currentPoint.y)) {
                addHighlight(currentPoint);
            }
        }

        /**
         * @return the highlighted
         */
        public LinkedList<Point2> getHighlighted() {
            return highlighted;
        }
    }
}